import time
import os
from typing import Optional

from audiomind.prompts import PROMPT_TEMPLATES

from audiomind.util import get_args, load_config, check, person_info, get_env_var
from audiomind.helpers.llm_util import initialize_llm
from audiomind.helpers.audio_utils import transcribe_and_store
from audiomind.helpers.summarize import summarize_text


class AudioMind:

    def __init__(self,
                 file: Optional[str] = None,
                 transcript_dir: Optional[str] = None,
                 whisper_model: Optional[str] = None,
                 openai_model: Optional[str] = None,
                 force_transcript: Optional[bool] = None,
                 whisper_api: Optional[bool] = None):
        """
            Initializes the AudioMind object.
            :param file: Path to the audio file.
            :param transcript_dir: Directory for transcripts.
            :param whisper_model: Whisper model for transcription.
            :param openai_model: OpenAI model for summarization.
            :param force_transcript: Flag to force transcription.
            :param whisper_api: Flag to use Whisper API.
        """
        self.load_configuration()
        args = get_args()
        self.file = self.set_env_if_value("AUDIO_FILE", file or args.file)
        self.transcript_dir = self.set_env_if_value("TRANSCRIPT_DIR", transcript_dir or args.transcript_dir)
        self.whisper_model = self.set_env_if_value("WHISPER_MODEL", whisper_model or args.whisper_model)
        self.openai_model = self.set_env_if_value("OPENAI_MODEL", openai_model or args.openai_model)
        self.force_transcript = self.set_env_if_value("FORCE_TRANSCRIPT", force_transcript or args.force_transcript)
        self.whisper_api = self.set_env_if_value("USE_WHISPER_API", whisper_api or args.whisper_api)
        check()
        self.audio_file = None
        self.base_name = None
        self.llm = None
        self.person_details = None
        self.transcript = None
        self.summary = None
        self.title_description = None
        self.setup()


    @classmethod
    def set_env_if_value(cls, env_var, value):
        if value is not None:
            os.environ[env_var] = value
        else:
            value = get_env_var(env_var)
        return value

    def load_configuration(self):
        """
        Loads configurations for AudioMind.
        """
        load_config()

    @staticmethod
    def get_main_file_name(file_name):
        main_file = os.path.splitext(os.path.basename(file_name))[0]
        return main_file

    def setup(self):
        self.llm = initialize_llm()
        self.person_details = person_info()
        self.audio_file = f"./{self.file}"
        self.base_name = self.get_main_file_name(self.file)

    def process(self):
        print('Starting...')
        try:
            start_time = time.time()
            self.transcribe()
            self.summarize()
            result_filename = self.save_results()
            end_time = time.time()
            elapsed = end_time - start_time
            print(f"Done in {elapsed:.2f} secs")
            print(f"Wrote results to {result_filename}")
        except Exception as e:
            print("Error:", e)

    def transcribe(self):
        start_time = time.time()
        print("Started transcribing...")
        self.transcript = transcribe_and_store(self.audio_file, self.base_name)
        end_time = time.time()
        elapsed_time = end_time - start_time
        print(f"Transcribing Done. Time taken: {elapsed_time:.2f} seconds")

    def summarize(self):
        start_time = time.time()
        print("Started summarizing...")
        local_start_time = time.time()
        summary_prompt_template = PROMPT_TEMPLATES["notes_template"]
        self.summary = summarize_text(self.llm, self.transcript, self.person_details, summary_prompt_template)

        elapsed_time = time.time() - local_start_time
        print(f"Generated Summary. Time taken: {elapsed_time:.2f}s")

        local_start_time = time.time()
        title_prompt_template = PROMPT_TEMPLATES["title_description_template"]
        self.title_description = summarize_text(self.llm, self.summary, self.person_details, title_prompt_template,
                                           chain_type="map_reduce",variables=["text"])

        elapsed_time = time.time() - local_start_time
        print(f"Generated Title. Time taken: {elapsed_time:.2f}s")

        final_elapsed_time = time.time() - start_time
        print(f"Summarizing Done. Time taken: {final_elapsed_time:.2f} seconds")

    def save_results(self):
        results_folder = get_env_var("RESULTS_DIR")
        os.makedirs(results_folder, exist_ok=True)
        filename = f"{results_folder}/{int(time.time())}_{self.base_name}.md"
        with open(filename, "w") as f:
            f.write(f"{self.title_description}\n\n{self.summary}\n")
        return filename



import os
from audiomind.util import get_env_var
import errno

from langchain.chat_models import ChatOpenAI
from langchain.document_loaders import TextLoader
from langchain.text_splitter import RecursiveCharacterTextSplitter


def initialize_llm():
    """Initialize the language model"""
    OPENAI_KEY = get_env_var("OPENAI_API_KEY")
    DREAMBOAT_KEY = get_env_var("DREAMBOAT_API_KEY")
    openai_model = get_env_var("OPENAI_MODEL")
    db_base_url = get_env_var("DREAMBOAT_BASE_URL")

    kwargs = {
        "temperature": 1,
        "model_name": openai_model,
        "openai_api_key": OPENAI_KEY
    }
    if DREAMBOAT_KEY:
        kwargs.update({
            "openai_api_base": db_base_url,
            "model_kwargs": {
                "headers": {
                    "x-dreamboat-api-key": DREAMBOAT_KEY,
                    "x-dreamboat-mode": "proxy openai"
                }
            }
        })
    return ChatOpenAI(**kwargs)


def load_transcript(transcript_file):
    if not os.path.exists(transcript_file):
        print("Error: Transcript file does not exist.")
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), transcript_file)
    return TextLoader(transcript_file).load()


def get_text_splitter():
    text_splitter = RecursiveCharacterTextSplitter(chunk_size=6000, chunk_overlap=20, length_function=len)
    return text_splitter



from dotenv import load_dotenv
import openai
import os
import argparse
import yaml
import pkg_resources


def get_env_var(var_name, check_exists=False):
    """Fetch and validate an environment variable"""
    var_value = os.getenv(var_name)
    if check_exists and not var_value:
        raise EnvironmentError(f"{var_name} is not set")
    if var_value is None and check_exists:
        raise EnvironmentError(f"{var_name} is not set")
    return var_value


def person_info(person_file=None):
    try:
        if person_file is None:
            person_file = get_env_var("USER_FILE")
        if not os.path.exists(person_file):
            print("Error: Person file does not exist.")
            return None
        with open(person_file, "r") as f:
            return f.read()
    except Exception as e:
        print("Error: ", e)
        return None


def get_args():
    parser = argparse.ArgumentParser(description='Process some flags for the script.')
    parser.add_argument('--file', help='The main file name to process')
    parser.add_argument('--transcript_dir', help='The directory where transcripts will be saved')
    parser.add_argument('--whisper_model', help='The whisper model to use')
    parser.add_argument('--openai_model', help='The AI model to use')
    parser.add_argument('--force_transcript', help='Force transcript generation')
    parser.add_argument('--whisper_api', help='Use the whisper API')
    args = parser.parse_args()
    return args


def load_yaml_config(file_path):
    with open(file_path, 'r') as file:
        return yaml.safe_load(file)


def load_config():
    load_dotenv()
    setup_openai()
    try:
        config_path = pkg_resources.resource_filename(__name__, 'helpers/config.yaml')
        config = load_yaml_config(config_path)
        for key, value in config.items():
            os.environ.setdefault(key, value)
    except Exception as e:
        print(e)
        print("Error loading config file.")


def setup_openai():
    openai.api_key = get_env_var("OPENAI_API_KEY")


def check():
    main_envs = ["AUDIO_FILE", "TRANSCRIPT_DIR", "WHISPER_MODEL", "OPENAI_MODEL", "RESULTS_DIR"]
    return [env for env in main_envs if get_env_var(env, check_exists=True) is None]



import time

from langchain.chains.summarize import load_summarize_chain
from langchain.docstore.document import Document
from langchain.prompts import PromptTemplate

from audiomind.prompts import PROMPT_TEMPLATES

from audiomind.helpers.llm_util import get_text_splitter


def make_docs(plain_text: str) -> list:
    texts = get_text_splitter().split_text(plain_text)
    docs = [Document(page_content=t) for t in texts]
    return docs


def summarize_docs(
    docs: list,
    prompt_template: str,
    details: str,
    model,
    variables: list,
    chain_type="stuff",
) -> str:
    prompt = PromptTemplate(template=prompt_template, input_variables=variables)

    if chain_type == "map_reduce":
        chain = load_summarize_chain(
            model, chain_type=chain_type, map_prompt=prompt, combine_prompt=prompt
        )
    else:
        chain = load_summarize_chain(model, chain_type=chain_type, prompt=prompt)
    values = {"input_documents": docs}
    if "details" in variables:
        values["details"] = details
    chain_output = chain(values, return_only_outputs=True)
    return chain_output["output_text"]


def summarize_text(
        llm, message: str, person_details: str, prompt_template: str,
        chain_type: str = "stuff", variables: list = ["text", "details"]
        ) -> str:
    docs = make_docs(message)
    summary_text = summarize_docs(
        docs,
        prompt_template,
        person_details,
        chain_type=chain_type,
        model=llm,
        variables=variables,
    )
    return summary_text



from setuptools import setup, find_packages

setup(
    name='audiomind',
    include_package_data=True,
    version='0.0.1',
    packages=find_packages(),
    install_requires=[
        'openai',
        'tiktoken',
        'openai-whisper',
        'langchain',
        'python-dotenv',
        'pydub',
        'pyyaml',
    ],
    entry_points={
        'console_scripts': [
            'audiomind=audiomind.main:main'
        ]
    }
)



PROMPT_TEMPLATES = {
    "summary_template": """Based on the provided transcript which is unstructured and chaotic, generate a detailed recap in bullet-point format. 
    Capture all the topics discussed exhaustively. 
    Group the bullet points into sections only if the topics naturally fall under a specific category. 
    If possible, include examples to highlight key points discussed.
Transcript:

{text}

Use Markdown. Add ## for section titles.""",

"notes_template": """
You are an expert copywriter. People gives you transcripts of their voice notes and you have to convert them into meaningful notes.
The voice note might be a journal entry, a meeting, a lecture, a podcast, etc. 
It is very unstructured and chaotic. 

I will give you the transcript of the voice note under [Transcript]
and details about the person who recorded it under [Person].

The reader is the same person, so you can use the details to personalize the notes but it is not necessary.
Dont unnecassarily use the person details in the notes.
Keep info in first person from the perspective of the person who recorded the voice note if possible and appropriate.
Transform the transcript into concise and structured notes, 
highlighting the key points and breakthroughs discussed.

Follow the chain of thought of the speaker and capture all the topics discussed exhaustively.
Dont make any assumptions. Dont add any extra information in the notes.

Dont make key points a long list of bullet points only. 
Group the thoughts or transcript into sections, and then add bullet points under each section.
Keep the order of the sections same as the order of the topics discussed in the transcript or their importance.

Include all action items and reflections 
Also, Create single section for action items / to-dos form the transcript and add them at the end of the notes.
Make sure to organize the information logically and eliminate any repetitive or irrelevant details.

[Transcript]
{text}
[Person]
{details}

Dont add any extra information in the notes.
I will create a title for journal myself. Dont add a heading / title in the notes.
Use Markdown. Add ## for section titles.
""",

"meeting_template": """
You are an expert copywriter. People gives you transcripts of their voice notes and you have to convert them into meaningful notes.
The voice note might be a journal entry, a meeting, a lecture, a podcast, etc. 
It is very unstructured and chaotic. 

I will give you the transcript of the meeting under [Transcript]
and details about the person who recorded it under [Person].

Dont unnecassarily use the person details in the notes.
Transform the transcript into concise and structured notes, 
highlighting the key points and breakthroughs discussed.

Follow the chain of thought of the speaker and capture all the topics discussed exhaustively.
Dont make any assumptions. Dont add any extra information in the notes.

Dont make key points a long list of bullet points only. 
Group the thoughts or transcript into sections, and then add bullet points under each section.
Keep the order of the sections same as the order of the topics discussed in the transcript or their importance.

Include all action items and reflections at the end of the notes..
Make sure to organize the information logically and eliminate any repetitive or irrelevant details.

[Transcript]
{text}
[Person]
{details}

Dont add any extra information in the notes.
I will create a title for journal myself. Dont add a heading / title in the notes.
Use Markdown. Add ## for section titles.
""",

    "title_description_template": """
I will give you notes in the form of a summary. It can be a journal entry, meeting notes or something else.
notes are divided into sections with bullet points with every section being a key point discussed in the notes.
Finally, it can consist of a list of action items / to-dos. Dont give action items and to-dos any priority in the title and description.
Your job is to generate single title and description for the notes.
write a short title and description, separate both by \n\n
Keep the title as short as possible, and make it a single line. Keep title below 100 characters.
The description should be a single paragraph. Keep description below 250 characters.

Text:
{text}
        
Return a Markdown in below format:
# TITLE HERE
#### DESCRIPTION HERE
""",
}



import os
import errno
import whisper
import json
import openai
from pydub.utils import mediainfo

from audiomind.util import get_env_var


def get_audio_length(audio_file):
    info = mediainfo(audio_file)
    length = float(info["duration"])
    return length


def calculate_transcription_cost(audio_file):
    audio_length = get_audio_length(audio_file)
    cost_per_second = 0.006 / 60
    total_cost = audio_length * cost_per_second
    decimal_places = len(str(cost_per_second).split('.')[1])
    print(f"Length of the audio file: {audio_length:.2f} seconds")
    print(f"Total transcription cost: ${total_cost:.{decimal_places}f}")
    return total_cost


def transcribe_and_store(audio_file, base_name):
    transcript_dir = get_env_var("TRANSCRIPT_DIR")
    os.makedirs(transcript_dir, exist_ok=True)
    transcript_file = f"./{transcript_dir}/{base_name}.txt"
    transcript = transcribe(audio_file, transcript_file)
    try:
        with open(transcript_file, "w") as f:
            if isinstance(transcript, dict):
                json.dump(transcript, f)
            else:
                f.write(transcript)
        return transcript
    except Exception as e:
        print("Error: ", e)
        return transcript


def transcribe(audio_file, transcript_file):
    if get_env_var("FORCE_TRANSCRIPT") != "1":
        if os.path.exists(transcript_file):
            with open(transcript_file, "r") as f:
                transcript = f.read()
                if len(transcript) > 0:
                    print("Transcript found in file. Skipping transcription.")
                    return transcript
    if not os.path.exists(audio_file):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), audio_file)

    if get_env_var("USE_WHISPER_API") == "1":
        return transcribe_openai_api(audio_file)
    result = transcribe_whisper_ondevice(audio_file)
    return result


def transcribe_openai_api(audio_file):
    print("Using OpenAI API to transcribe audio file.")
    calculate_transcription_cost(audio_file)
    audio = open(audio_file, "rb")
    transcript = openai.Audio.transcribe("whisper-1", audio)
    result = transcript["text"]
    return result


def transcribe_whisper_ondevice(audio_file):
    print("Using Local Whisper to transcribe audio file.")
    whisper_model = get_env_var("WHISPER_MODEL")
    model = whisper.load_model(whisper_model, in_memory=True)
    a = model.transcribe(audio_file, verbose=False)
    result = a["text"]
    return result

